# -*- coding: utf-8 -*-
"""
Stand-in extension of Qt's QMutex and QRecursiveMutex classes.
Derived from the ACQ4 project.

Copyright (c) 2010, Luke Campagnola.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['Mutex', 'RecursiveMutex']

from PySide2.QtCore import QMutex as _QMutex
from PySide2.QtCore import QRecursiveMutex as _QRecursiveMutex
from typing import Optional, Union


_RealNumber = Union[int, float]


class Mutex(_QMutex):
    """ Extends QMutex which serves as access serialization between threads.

    This class provides:
    * Drop-in replacement for threading.Lock
    * Context management (enter/exit)
    """

    def acquire(self, blocking: Optional[bool] = True, timeout: Optional[_RealNumber] = -1) -> bool:
        """ Mimics threading.Lock.acquire() to allow this class as a drop-in replacement.

        @param bool blocking: If True, this method will be blocking (indefinitely or for up to
                              <timeout> seconds) until the mutex has been locked. If False this
                              method will return immediately independent of the ability to lock.
        @param float timeout: Timeout in seconds specifying the maximum wait time for the mutex to
                              be able to lock. Negative numbers correspond to infinite wait time.
                              This parameter is ignored for <blocking> == False.
        """
        if blocking:
            # Convert to milliseconds for QMutex
            return self.tryLock(max(-1, int(timeout * 1000)))
        return self.tryLock()

    def release(self) -> None:
        """ Mimics threading.Lock.release() to allow this class as a drop-in replacement.
        """
        self.unlock()

    def __enter__(self):
        """ Enter context.

        @return Mutex: this mutex
        """
        self.lock()
        return self

    def __exit__(self, *args):
        """ Exit context.

        @param args: context arguments (type, value, traceback)
        """
        self.unlock()


# Compatibility workaround for PySide2 vs. PySide6. In PySide2 we need to use QMutex with an
# initializer argument to construct a recursive mutex but in PySide6 we need to subclass
# QRecursiveMutex. Check if QRecursiveMutex class has all API members (indicating it's PySide6).
if all(hasattr(_QRecursiveMutex, attr) for attr in ('lock', 'unlock', 'tryLock')):
    class RecursiveMutex(_QRecursiveMutex):
        """ Extends QRecursiveMutex which serves as access serialization between threads.

        This class provides:
        * Drop-in replacement for threading.Lock
        * Context management (enter/exit)

        NOTE: A recursive mutex is much more expensive than using a regular mutex. So consider
        refactoring your code to use a simple mutex before using this object.
        """

        def acquire(self, blocking: Optional[bool] = True,
                    timeout: Optional[_RealNumber] = -1) -> bool:
            """ Mimics threading.Lock.acquire() to allow this class as a drop-in replacement.

            @param bool blocking: If True, this method will be blocking (indefinitely or for up to
                                  <timeout> seconds) until the mutex has been locked. If False this
                                  method will return immediately independent of the ability to lock.
            @param float timeout: Timeout in seconds specifying the maximum wait time for the mutex to
                                  be able to lock. Negative numbers correspond to infinite wait time.
                                  This parameter is ignored for <blocking> == False.
            """
            if blocking:
                # Convert to milliseconds for QMutex
                return self.tryLock(max(-1, int(timeout * 1000)))
            return self.tryLock()

        def release(self) -> None:
            """ Mimics threading.Lock.release() to allow this class as a drop-in replacement.
            """
            self.unlock()

        def __enter__(self):
            """ Enter context.

            @return RecursiveMutex: this mutex
            """
            self.lock()
            return self

        def __exit__(self, *args):
            """ Exit context.

            @param args: context arguments (type, value, traceback)
            """
            self.unlock()
else:
    class RecursiveMutex(Mutex):
        """ Extends QRecursiveMutex which serves as access serialization between threads.

        This class provides:
        * Drop-in replacement for threading.Lock
        * Context management (enter/exit)

        NOTE: A recursive mutex is much more expensive than using a regular mutex. So consider
        refactoring your code to use a simple mutex before using this object.
        """
        def __init__(self):
            super().__init__(_QMutex.Recursive)



# -*- coding: utf-8 -*-
"""
This file contains qudi methods for mathematical operations/transformations.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ('compute_ft', 'ft_windows')

import numpy as np
from scipy import signal


# Available windows to be applied on signal data before FT.
# To find out the amplitude normalization factor check either the scipy implementation on
#     https://github.com/scipy/scipy/blob/v0.15.1/scipy/signal/windows.py#L336
# or just perform a sum of the window (oscillating parts of the window should be averaged out and
# constant offset factor will remain):
#     MM=1000000  # choose a big number
#     print(sum(signal.hanning(MM))/MM)
ft_windows = {'none': {'func': np.ones, 'ampl_norm': 1.0},
              'hamming': {'func': signal.hamming, 'ampl_norm': 1.0/0.54},
              'hann': {'func': signal.hann, 'ampl_norm': 1.0/0.5},
              'blackman': {'func': signal.blackman, 'ampl_norm': 1.0/0.42},
              'triang': {'func': signal.triang, 'ampl_norm': 1.0/0.5},
              'flattop': {'func': signal.flattop, 'ampl_norm': 1.0/0.2156},
              'bartlett': {'func': signal.bartlett, 'ampl_norm': 1.0/0.5},
              'parzen': {'func': signal.parzen, 'ampl_norm': 1.0/0.375},
              'bohman': {'func': signal.bohman, 'ampl_norm': 1.0/0.4052847},
              'blackmanharris': {'func': signal.blackmanharris, 'ampl_norm': 1.0/0.35875},
              'nuttall': {'func': signal.nuttall, 'ampl_norm': 1.0/0.3635819},
              'barthann': {'func': signal.barthann, 'ampl_norm': 1.0/0.5}
              }


def compute_ft(x_val, y_val, zeropad_num=0, window='none', base_corr=True, psd=False):
    """Compute the Discrete fourier Transform of the power spectral density

    @param numpy.array x_val: 1D array
    @param numpy.array y_val: 1D array of same size as x_val
    @param int zeropad_num: optional, zeropadding (adding zeros to the end of the array).
                            zeropad_num >= 0, the size of the array which is add to the end of the
                            y_val before performing the Fourier Transform (FT). The resulting array
                            will have the length (len(y_val)/2)*(zeropad_num+1)
                            Note that zeropadding will not change or add more information to the
                            dft, it will solely interpolate between the dft_y values (corresponds
                            to a Sinc interpolation method).
                            Set zeropad_num=1 to obtain output arrays which have the same size as
                            the input arrays. Default is zeropad_num=0.
    @param str window: optional, the window function which should be applied to the y values before
                       Fourier Transform is calculated.
    @param bool base_corr: optional, select whether baseline correction shoud be performed before
                           calculating the FT. Default is False.
    @param bool psd: optional, select whether the Discrete Fourier Transform or the Power Spectral
                     Density (PSD, which is just the FT of the absolute square of the y-values)
                     should be computed. Default is psd=False.

    @return: tuple(dft_x, dft_y): be aware that the return arrays' length depend on the zeropad_num
                                  like len(dft_x) = len(dft_y) = (len(y_val)/2)*(zeropad_num+1)

    Pay attention that the return values of the FT have only half of the entries compared to the
    used signal input (if zeropad=0).

    In general, a window function should be applied on the y data before calculating the FT, to
    reduce spectral leakage. The Hann window for instance is almost never a bad choice. Use it like
    window='hann'

    Keep always in mind the relation to the Fourier transform space:
        T = delta_t * N_samples
    where delta_t is the distance between the time points and N_samples are the amount of points in
    the time domain. Consequently the sample rate is:
        f_samplerate = T / N_samples

    Keep also in mind that the FT returns value from 0 to f_samplerate, or equivalently
    -f_samplerate/2 to f_samplerate/2.

    Difference between PSD and DFT:
    The power spectral density (PSD) describes how the power of your signal is distributed over
    frequency whilst the DFT shows the spectral content of your signal, i.e. the amplitude and
    phase of harmonics in your signal.
    """

    x_val = np.array(x_val)
    y_val = np.array(y_val)

    # Make a baseline correction to avoid a constant offset near zero frequencies. Offset of the
    # y_val from mean corresponds to half the value at fft_y[0]
    corrected_y = y_val
    if base_corr:
        corrected_y = y_val - y_val.mean()

    ampl_norm_fact = 1.0
    # apply window to data to account for spectral leakage:
    if window in ft_windows:
        window_val = ft_windows[window]['func'](len(y_val))
        corrected_y = corrected_y * window_val
        # to get the correct amplitude in the amplitude spectrum
        ampl_norm_fact = ft_windows[window]['ampl_norm']

    # zeropad for sinc interpolation:
    zeropad_arr = np.zeros(len(corrected_y)*(zeropad_num+1))
    zeropad_arr[:len(corrected_y)] = corrected_y

    # Get the amplitude values from the fourier transformed y values.
    fft_y = np.abs(np.fft.fft(zeropad_arr))

    # Power spectral density (PSD) or just amplitude spectrum of fourier signal:
    power_value = 1.0
    if psd:
        power_value = 2.0

    # The factor 2 accounts for the fact that just the half of the spectrum was taken. The
    # ampl_norm_fact is the normalization factor due to the applied window function (the offset
    # value in the window function):
    fft_y = ((2/len(y_val)) * fft_y * ampl_norm_fact)**power_value

    # Due to the sampling theorem you can only identify frequencies at half of the sample rate,
    # therefore the FT contains an almost symmetric spectrum (the asymmetry results from aliasing
    # effects). Therefore take the half of the values for the display.
    middle = int((len(zeropad_arr)+1)//2)

    # sample spacing of x_axis, if x is a time axis than it corresponds to a timestep:
    x_spacing = np.round(x_val[-1] - x_val[-2], 12)

    # use the helper function of numpy to calculate the x_values for the fourier space. That
    # function will handle an occuring devision by 0:
    fft_x = np.fft.fftfreq(len(zeropad_arr), d=x_spacing)
    return abs(fft_x[:middle]), fft_y[:middle]



# -*- coding: utf-8 -*-

import sys
from setuptools import setup, find_namespace_packages


unix_dep = [
    'wheel',
    'cycler',
    'entrypoints',
    'fysom',
    'GitPython',
    'jupyter',
    'jupytext',
    'lmfit>=1.0.3',
    'matplotlib',
    'numpy',
    'pyqtgraph>=0.13.0',
    'PySide2==5.15.2',
    'rpyc>=5.0.1',
    'ruamel.yaml>=0.17.16',
    'scipy>=1.7.1',
    'jsonschema>=4.2.1',
]

windows_dep = [
    'wheel>=0.37.0',
    'cycler>=0.10.0',
    'entrypoints>=0.3',
    'fysom>=2.1.6',
    'GitPython>=3.1.24',
    'jupyter>=1.0.0',
    'jupytext>=1.13.0',
    'lmfit>=1.0.3',
    'matplotlib>=3.4.3',
    'numpy>=1.21.3',
    'pyqtgraph>=0.13.0',
    'PySide2==5.15.2',
    'rpyc>=5.0.1',
    'ruamel.yaml>=0.17.16',
    'scipy>=1.7.1',
    'jsonschema>=4.2.1',
]

with open('VERSION', 'r') as file:
    version = file.read().strip()

with open('README.md', 'r') as file:
    long_description = file.read()

setup(
    name='qudi-core',
    version=version,
    packages=find_namespace_packages(where='src', exclude=['qudi.artwork']),
    package_dir={'': 'src'},
    package_data={'': ['LICENSE', 'LICENSE.LESSER', 'AUTHORS.md', 'README.md', 'VERSION'],
                  'qudi': ['artwork/icons/*',
                           'artwork/icons/**/*',
                           'artwork/icons/**/**/*',
                           'artwork/styles/*',
                           'artwork/styles/**/*',
                           'artwork/styles/**/**/*',
                           'artwork/logo/*',
                           ],
                  },
    description='A modular measurement application framework',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Ulm-IQO/qudi-core',
    project_urls={'Documentation': 'https://ulm-iqo.github.io/qudi-core/',
                  'Source Code': 'https://github.com/Ulm-IQO/qudi-core/',
                  'Bug Tracker': 'https://github.com/Ulm-IQO/qudi-core/issues/',
                  },
    keywords=['qudi',
              'diamond',
              'quantum',
              'confocal',
              'experiment',
              'measurement',
              'framework',
              'lab',
              'laboratory',
              'instrumentation',
              'instrument',
              'modular'
              ],
    license='LGPLv3',
    install_requires=windows_dep if sys.platform == 'win32' else unix_dep,
    python_requires='>=3.8, <3.10',
    classifiers=['Development Status :: 5 - Production/Stable',

                 'Environment :: Win32 (MS Windows)',
                 'Environment :: X11 Applications',
                 'Environment :: MacOS X',

                 'Intended Audience :: Developers',
                 'Intended Audience :: Science/Research',
                 'Intended Audience :: End Users/Desktop',

                 'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

                 'Natural Language :: English',

                 'Operating System :: Microsoft :: Windows :: Windows 8',
                 'Operating System :: Microsoft :: Windows :: Windows 8.1',
                 'Operating System :: Microsoft :: Windows :: Windows 10',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: Unix',
                 'Operating System :: POSIX :: Linux',

                 'Programming Language :: Python :: 3.8',
                 'Programming Language :: Python :: 3.9',

                 'Topic :: Scientific/Engineering',
                 'Topic :: Software Development :: Libraries :: Application Frameworks',
                 'Topic :: Software Development :: User Interfaces',
                 ],
    entry_points={
        'console_scripts': ['qudi=qudi.runnable:main',
                            'qudi-config-editor=qudi.tools.config_editor.config_editor:main',
                            'qudi-uninstall-kernel=qudi.core.qudikernel:uninstall_kernel',
                            'qudi-install-kernel=qudi.core.qudikernel:install_kernel'
                            ]
    },
    zip_safe=False
)



# -*- coding: utf-8 -*-
"""
This file contains Qudi methods for handling real-world values with units.

Copyright (c) 2021, the qudi developers. See the AUTHORS.md file at the top-level directory of this
distribution and on <https://github.com/Ulm-IQO/qudi-core/>

This file is part of qudi.

Qudi is free software: you can redistribute it and/or modify it under the terms of
the GNU Lesser General Public License as published by the Free Software Foundation,
either version 3 of the License, or (at your option) any later version.

Qudi is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with qudi.
If not, see <https://www.gnu.org/licenses/>.
"""

__all__ = ['ScalarConstraint']

from typing import Union, Optional, Tuple, Callable, Any
from qudi.util.helpers import is_float, is_integer


class ScalarConstraint:
    """
    """
    def __init__(self,
                 default: Union[int, float],
                 bounds: Tuple[Union[int, float], Union[int, float]],
                 increment: Optional[Union[int, float]] = None,
                 enforce_int: Optional[bool] = False,
                 checker: Optional[Callable[[Union[int, float]], bool]] = None
                 ) -> None:
        """
        """
        self._enforce_int = bool(enforce_int)
        self._check_value_type(default)
        for value in bounds:
            self._check_value_type(value)
        if increment is not None:
            self._check_value_type(increment)
        if checker is not None and not callable(checker):
            raise TypeError('checker must be eithe None or a callable accepting a single scalar '
                            'and returning a bool.')
        self._default = default
        self._minimum, self._maximum = sorted(bounds)
        self._increment = increment
        self._checker = checker

        if not self.is_valid(self._default):
            raise ValueError(f'invalid default value ({self._default}) encountered')

    @property
    def bounds(self) -> Tuple[Union[int, float], Union[int, float]]:
        return self._minimum, self._maximum

    @property
    def minimum(self) -> Union[int, float]:
        return self._minimum

    @property
    def maximum(self) -> Union[int, float]:
        return self._maximum

    @property
    def default(self) -> Union[int, float]:
        return self._default

    @property
    def increment(self) -> Union[None, int, float]:
        return self._increment

    @property
    def enforce_int(self) -> bool:
        return self._enforce_int

    def is_valid(self, value: Union[int, float]) -> bool:
        try:
            self._check_value_type(value)
        except TypeError:
            return False

        if self._minimum <= value <= self._maximum:
            if self._checker is None:
                return True
            else:
                return self._checker(value)
        return False

    def clip(self, value: Union[int, float]) -> Union[int, float]:
        return min(self._maximum, max(self._minimum, value))

    def copy(self) -> object:
        return ScalarConstraint(default=self.default,
                                bounds=self.bounds,
                                increment=self.increment,
                                enforce_int=self.enforce_int,
                                checker=self._checker)

    def _check_value_type(self, value: Any) -> None:
        if self._enforce_int:
            if not is_integer(value):
                raise TypeError(f'values must be int type (received {value})')
        else:
            if not (is_integer(value) or is_float(value)):
                raise TypeError(f'values must be int or float type (received {value})')

    def __repr__(self) -> str:
        cls = self.__class__.__name__
        module = self.__class__.__module__
        return f'{module}.{cls}(' \
               f'default={self.default}, ' \
               f'bounds={self.bounds}, ' \
               f'increment={self.increment}, ' \
               f'enforce_int={self.enforce_int}, ' \
               f'checker={self._checker})'

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memodict={}):
        new_obj = self.copy()
        memodict[id(self)] = new_obj
        return new_obj

    # Backwards compatibility properties:
    @default.setter
    def default(self, value: Union[int, float]):
        if not self.is_valid(value):
            raise ValueError(f'invalid default value ({value}) encountered')
        self._default = value

    @property
    def min(self) -> Union[int, float]:
        return self._minimum

    @min.setter
    def min(self, value: Union[int, float]):
        self._minimum = value

    @property
    def max(self) -> Union[int, float]:
        return self._maximum

    @max.setter
    def max(self, value: Union[int, float]):
        self._maximum = value

    @property
    def step(self) -> Union[None, int, float]:
        return self._increment

    @step.setter
    def step(self, value: Union[None, int, float]):
        self._increment = value
